/*
    libporty - async background processes, threaded version
    Copyright (C) 2010..2012  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page: http://repo.hu/projects/libporty
    Author: libporty@igor2.repo.hu
*/

#include <libporty/bgtask/bgtask.h>
#include <libporty/host/pointers.h>
#include <libporty/net/uninit_chain.h>
#include <assert.h>


/* Watchdog is a thread watching for background processes and report events
   on control sockets */
static int watchdogt = -1;
static P_net_socket watchdogs = P_NET_INVALID_SOCKET;

/* wait for numpid processes, return the index of the one that exited or -1 if all running */
static int watchdog_wait(P_net_socket cfd, int numpids, P_pid_t *pids, bgproc_status_t *exit_status);

/* kill a process */
static void watchdog_kill(P_pid_t pid, int hard);

static void P_watchdog_process(P_net_socket fd, int argc, char *argv[], P_pid_t pid)
{
	if (strcmp(argv[0], "TERM") == 0) {
		watchdog_kill(pid, 0);
	}
	else if (strcmp(argv[0], "KILL") == 0) {
		watchdog_kill(pid, 1);
	}
}

static unsigned long int children_created = 0, children_seen = 0;

static int watchdog(void *arg)
{
	P_net_socket *cfd = (P_net_socket *)arg;
	P_buf_t **buf = NULL;
	P_pid_t *pids = NULL;
	int len, n, exited;
	char rb[256];
	char *request;
	struct P_pollfd *fds = NULL;
	P_net_socket *chfds = NULL;
	int alloced = 0, used = 0;
	char *argv[4];
	int argc;
	bgproc_status_t st;
	int children_active;

#define addfd(fd_, events_, pid_, chfd_) \
	do { \
		if (used >= alloced) { \
			alloced += 16; \
			fds   = realloc(fds, alloced * sizeof(struct P_pollfd)); \
			chfds = realloc(chfds, alloced * sizeof(P_net_socket)); \
			buf   = realloc(buf, alloced * sizeof(P_buf_t *)); \
			pids  = realloc(pids, alloced * sizeof(P_pid_t)); \
		} \
		chfds[used] = chfd_; \
		fds[used].fd = fd_; \
		fds[used].events = events_; \
		fds[used].revents = 0; \
		buf[used] = P_buf_init("\r\n"); \
		pids[used] = (pid_); \
		used++; \
	} while(0)

#define delfd(idx) \
	do { \
		if (idx < used - 1) { \
			int len; \
			len = used - idx - 1; \
			memmove(fds + idx,  fds + idx + 1,  len * sizeof(struct P_pollfd)); \
			memmove(chfds + idx,  chfds + idx + 1,  len * sizeof(P_net_socket)); \
			memmove(buf + idx,  buf + idx + 1,  len * sizeof(P_buf_t *)); \
			memmove(pids + idx, pids + idx + 1, len * sizeof(P_pid_t)); \
		} \
		used--; \
	} while(0)

	addfd(*cfd, P_POLLIN, P_NULL_PID, P_NET_INVALID_SOCKET);

	children_active = 0;
	for(;;) {
		if (children_active > 0) {
			exited = watchdog_wait(*cfd, used-1, pids+1, &st); /* wait for all pids except for 0, which is our control socket without pid */
			if (exited >= 0) {
				children_active--;
				exited++; /* compensate for our cheat to hide index 0 */
				len = sprintf(rb, "END %s %d\n", reasons[st.reason], st.code);
/*			printf("Sending to %d: %s\n", fds[exited].fd, rb);*/
				send(fds[exited].fd, rb, len, 0);

/*	fprintf(stderr, "Closing ctrlfd=%d chfd=%d id=%d\n", fds[exited].fd, chfds[exited], exited, used);*/
				if (chfds[exited] >= 0)
					P_net_close(chfds[exited]);
				if (fds[exited].fd >= 0)
					P_net_close(fds[exited].fd);
				P_buf_uninit(buf[exited]);
				delfd(exited);
				continue;
			}
		}
		P_poll(fds, used, 100);
		for(n = 0; n < used; n++) {
			if (fds[n].revents & P_POLLIN) {
				len = recv(fds[n].fd, rb, sizeof(rb)-1, 0);
/*	printf("watchdog len: %d on %d\n", len, n); */
				if (len > 0) {
					rb[len] = '\0';
					P_buf_fill(buf[n], rb);
					while((request = P_buf_get(buf[n])) != NULL) {
						P_argv_split(request, &argc, argv, 4, " \t");
						{
							if (n == 0) {
/*	printf("watchdog read: %s\n", argv[0]); */
								if (strcmp(argv[0], "N") == 0) {
									int *fd = NULL;
									P_pid_t *pid = NULL;
									int *chfd = NULL;
									assert(P_ptr_scan(argv[1], (void **)&fd) == 0);
									assert(P_ptr_scan(argv[2], (void **)&pid) == 0);
									assert(P_ptr_scan(argv[3], (void **)&chfd) == 0);
/*	fprintf(stderr, "WATCHDOG add: fd=%s/%p/%d pid=%s/%p/%d chfd=%s/%p/%d id=%d\n", argv[1], fd, *fd, argv[2], pid, *pid, argv[3], chfd, *chfd, used);*/
									children_active++;
									addfd(*fd, P_POLLIN, *pid, *chfd);
									free(fd);
									free(pid);
									free(chfd);
									children_seen++;
								}
							}
							else
								P_watchdog_process(fds[n].fd, argc, argv, pids[n]);
						}
						P_buf_free(request);
					}
				}
				else {
					if (n == 0)
						goto ctrl_eof;
#warning handle eof on other control sockets!
				}
			}
		}
	}

ctrl_eof:;

	for(n = 0; n < used; n++) {
		if (chfds[n] != -1) /* sometimes there's no valid chfd, for example when n == 0 (control socket) */
			P_net_close(chfds[n]);
		P_buf_uninit(buf[n]);
	}

	free(fds);
	free(chfds);
	free(buf);
	free(pids);

	P_net_close(*cfd);

	return 0;
#undef addfd
}

void watchdog_uninit(void)
{
	if (watchdogt >= 0) {
		P_net_close(watchdogs);
		watchdogt = -1;
	}
#warning TODO: do not sleep here
	P_usleep(10000);
}

static void watchdog_add(P_net_socket *ctrlfd, P_pid_t *pid, P_net_socket *chfd)
{
	static P_net_socket rfd, *cfd, *childfd;
	char buff[P_PTR_STRWIDTH * 3 + 8];
	int len;
	P_pid_t *p;
	char pbuff[3][P_PTR_STRWIDTH];

	if ((ctrlfd == NULL) || (pid == NULL))
		return;

	if (watchdogt < 0) {
		P_net_local_socket(&watchdogs, &rfd, P_net_nonblock_full);
		watchdogt = P_bgtask_create(watchdog, &rfd);
		P_net_uninit_reg(watchdog_uninit);
	}

	cfd = malloc(sizeof(P_net_socket));
	if (ctrlfd == NULL)
		memset(cfd, 0, sizeof(P_net_socket));
	else
		memcpy(cfd, ctrlfd, sizeof(P_net_socket));

	childfd = malloc(sizeof(P_net_socket));
	memcpy(childfd, chfd, sizeof(P_net_socket));

	p = malloc(sizeof(P_pid_t));
	memcpy(p, pid, sizeof(P_pid_t));

	P_ptr_print(pbuff[0], cfd);
	P_ptr_print(pbuff[1], p);
	P_ptr_print(pbuff[2], childfd);

	len = sprintf(buff, "N %s %s %s\n", pbuff[0], pbuff[1], pbuff[2]);

/*	printf("watchdog SEND: '%s'/%d fd=%d pid=%d\n", buff, len, *(int *)cfd, *(int *)pid);*/
#warning should check return value
	send(watchdogs, buff, len, 0);

	len = sprintf(buff, "START\n");
	send(*ctrlfd, buff, len, 0);
}

